

%% about this code

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
%           This code solves the dynamic optimization problem             %
%       described in the paper `The shadow costs of illiquidity'          % 
%      by Kristy A.E. Jansen and Bas J.M.Werker (Tilburg University)      %
%                                                                         %
%             In case of questions, contact Kristy Jansen:                %
%                k.a.e.jansen@tilburguniversity.edu                       %
%                                                                         %
%                  This version: December 2020.                           %
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

%% licenses
% matlab (MATLAB_R2019b)
% optimization toolbox 
% statistics toolbox
% curve fitting toolbox

%% important notes
% In case of extreme parameter choices (correlations close to 1, extreme 
% prices of risk, extreme liquidity shocks etc.), the choice of initial 
% parameters values and the options how to search for the optimal solution 
% should be carefully re-considered.
% Failure of the dynamic optimization problem is easily detected by
% evaluating the exitflag and/or sum of squared residuals at each grid
% search.

%% initialization
clear; clc;

%set simulations dynamic programming technique
M=10^5;                                                                     %number of scenarios 
iota=ones(M,1);                                                             %vector of ones

%parameter values
gamma=5;                                                                    %risk aversion parameter 
r_y=0.02;                                                                   %riskfree rate annually
lambda=0.38; sigma=0.185; mu=sigma*lambda+r_y;                              %risk premium  %standard deviation %expected return liquid risky asset 
lambda2=0.38; sigma2=0.185; mu2=sigma2*lambda2+r_y;                         %risk premium  %standard deviation %expected return illiquid risky asset 
rho=0;                                                                      %correlation between liquid risky asset and illiquid risky asset
beta=0.91;                                                                  %time preference
T=1;                                                                       %final date T
freq=12;                                                                    %trading frequency (per year)    

ll=0.3; prob_l=0.1;                                                         %liquidity shock parameters

tc=0.01;                                                                    %transaction costs
discount=tc;                                                                %discount at which the illiquid asset can be sold at final date T

p=0.5;                                                                       %intensity to trade

%return process
covar=rho*sigma*sigma2;                                                     %covariance between liquid risky asset and illiquid risky asset
varcovar_matrix=[sigma^2 covar; covar sigma2^2];                            %variance-covariance matrix risky assets
drift_geometric_prices=[mu-0.5*sigma^2 mu2-0.5*sigma2^2];                   %Black-Scholes world; drift of log-prices for the liquid and illiquid risky asset respectively

%dividend
prob_d=1; 
d=0.0;
D=d*binornd(1,prob_d,M,1);

%merton solution
mertonsol=(1/gamma)*(varcovar_matrix)^(-1)*([mu mu2]'-[r_y r_y]');          %solution to the merton two risky asset case


%% return, trading opportunity, and liquidity shock dynamics
%risk-free (h-frequency)
r_y_h=r_y/freq; 
r_exp=exp(r_y_h*iota);

%liquid and illiquid risky assets  (h-frequency)
R_riskyassets_normal=zeros(M,2,T*freq); R_riskyassets_exp=zeros(M,2,T*freq);  
for dn=1:T*freq
R_riskyassets_normal(:,:,dn)=mvnrnd(drift_geometric_prices/freq,varcovar_matrix/freq,M); %generating M-returns normal-distributed returns liquid and illiquid risky asset respectively
R_riskyassets_exp(:,:,dn)=exp(R_riskyassets_normal(:,:,dn));   %exponentials over returns such that wealth does not turn negative
end

%trading opportunities
p_t=zeros(M,T*freq);
for dn=1:T*freq
    p_t(:,dn)=binornd(1,1-exp(-p*1/freq),M,1); %probability to trade (h-frequency)
end   

%liquidity shock
q_t=zeros(M,T*freq); 
for dn=1:T*freq
    q_t(:,dn)=binornd(1,1-exp(-prob_l*1/freq),M,1); %probability liquidity shock (h-frequency)
end   
LI=ll*q_t; %liquidity shock  


%% dynamic programming initialization 
% set grid search
Numsteps=100; %grid fraction of illiquid wealth
Numsteps_delta=100; %grid for evaluation of shadow costs

%illiquid case specification 
%initialization H-functions
H_initial=1; H_NT_T=zeros(1,T*freq+1); H_NT_T(1,1)=H_initial; %boundary condition H-function when trade
H_NT=zeros(Numsteps,T*freq+1); H_der_NT=zeros(Numsteps, T*freq+1); %boundary condition H-function when no trade
H_NT_T_with_TC=zeros(Numsteps,T*freq+1); H_NT_T_with_TC_int=zeros(Numsteps,T*freq+1); %boundary condition H-function when trade + transaction costs
H_comb=zeros(M,Numsteps,T*freq+1); H_comb_der=zeros(M,Numsteps,T*freq+1);

%H-function final date 
for j=1:Numsteps                                                            
H_NT(j,1)=(1-discount*((j-1)/(Numsteps)))^(1-gamma); %discount at which the illiquid asset can be sold at T
end
H_NT_T_with_TC(:,1)=H_NT(:,1); H_NT_T_with_TC_int(:,1)=H_NT(:,1); 

% H-der function 
for j=1:Numsteps
    if j==1 
    H_der_NT(j,1)=(H_NT(j+1,1)-H_NT(j,1))/(1/Numsteps);
elseif j==Numsteps
        H_der_NT(j,1)=(H_NT(j,1)-H_NT(j-1,1))/(1/Numsteps);
    else
H_der_NT(j,1)=(H_NT(j+1,1)-H_NT(j-1,1))/(2/Numsteps);
    end
end

%combine H-function when trade and no trade (final date: always pay
%discount/transaction costs irrespecitve trading opportunity)
for j=1:Numsteps
    for i=1:M
    H_comb(i,j,1)=H_NT(j,1);
    H_comb_der(i,j,1)=H_der_NT(j,1);
    end
end

%liquid case specification 
H_T_liquid=zeros(Numsteps_delta-1,T*freq+1); %H-function for liquid case
H_T_liquid(:,1)=H_initial.^(1-gamma);


%% solve illiquid case
%solve model backwards: start the period before the final date and solve
%till we arrive at time t=0

%define vectors and matrices
xi=zeros(Numsteps,1); 
parameter_estimates=zeros(Numsteps, T*freq, 2); sum_square_residuals=zeros(Numsteps,T*freq); 
KKT_illiquid=zeros(Numsteps,T*freq,2); termination=zeros(Numsteps,T*freq); xi_ster_TC=zeros(T*freq,1); 
consumption_ster_TC=zeros(T*freq,1); liquid_ster_TC=zeros(T*freq,1); H_comb_ster_TC=zeros(M,T*freq); 

%set options optimization problem
options=optimoptions('lsqnonlin','FiniteDifferenceStepSize',1e-9,'TolFun',1.e-9,'MaxFunEvals',1000,'MaxIterations',1000);  

%define grid xi
for j=1:Numsteps
    xi(j,1)=(j-1)/Numsteps;
end
xi_adj=xi*ones(1,T*freq);

%start dynamic optimization problem
lb=[0,-inf]; %lowerbound optimization

if 1-exp(-p*1/freq)==1 %upper bound liquid case
ub = [1,inf];
end
    
for dn=1:T*freq  
    
parameter_initial=[1/(dn+1)*(1-ll) mertonsol(1,1)]; %initial paramter values 
    
    for j=1:Numsteps     
         %optimal solution if at boundary; derived in paper
         if j>1 && round(parameter_estimates(j-1,dn,1),3)==round(1-xi(j-1,1)*(1-d)-ll,3) || 1-xi(j,1)<=ll
         parameter_estimates(j,dn,:)=[max(1-xi(j,1)*(1-d)-ll,0.00001) 0];
         sum_square_residuals(j,dn)=NaN;
         termination(j,dn)=NaN;
         KKT_illiquid(j,dn,:)=[NaN NaN];
         else          
                
                 if 1-xi(j,1)*(1-d)-ll>0.0001 && 1-exp(-p*1/freq)<1
                    ub = [1-xi(j,1)*(1-d)-ll,inf]; %upper bound: only consume and pay liquidity shock out of liquid wealth, so alpha<1-xi-ll
                 end      

         fun_NT = @(x) focs_NT(x,xi(j,1),beta,gamma,r_exp,R_riskyassets_exp(:,:,dn),H_comb(:,j,dn),H_comb_der(:,j,dn),iota,LI(:,dn),D,1/freq); %identifying function to be solved

         [x0_notrade,resnorm,lambda,exitflag]=lsqnonlin(fun_NT,parameter_initial,lb,ub,options); %applying lsqnonlin to obtain optimal solution for alpha_xi and theta_xi
         parameter_estimates(j,dn,:)=x0_notrade; %solution for xi(j,1)=(j-1)/Numsteps at time T-dn*h
         parameter_initial=x0_notrade; %update initial parameters
         sum_square_residuals(j,dn)=resnorm; %obtain sum squared residuals
         termination(j,dn)=exitflag; %reason solver stopped
         KKT_illiquid(j,dn,:)=lambda; %KKT conditions
         end
    end
    
    for j=1:Numsteps
     %correct for unconverged solutions (typically only for high xi(j,1))
     %exitflag=0; number of iterations exceeds options.MaxIterations or number of function evaluations exceeded options.MaxFunctionEvaluations (source: Matlab)
     %exitflag=2; Jacobian could be undefined (source: Matlab)
     if j>3 && j<Numsteps-3 && termination(j,dn)==0 || j>3 && j<Numsteps-3 && termination(j,dn)==2
         for k=1:5
             ktemp=k-3;
             if ~isnan(termination(j+ktemp,dn))       
              parameter_estimates(j+ktemp,dn,:)=parameter_estimates(j-2,dn,:);
              xi_adj(j+ktemp,dn)=xi(j-2,1);
             end
         end
    end       
             
    %updating H when no trade
    H_NT(j,dn+1)=(parameter_estimates(j,dn,1))^(1-gamma)+beta^(1/freq)*mean((((iota-xi_adj(j,dn)*iota-parameter_estimates(j,dn,1)*iota-LI(:,dn)).*(r_exp+...
    parameter_estimates(j,dn,2)*(R_riskyassets_exp(:,1,dn)-r_exp))+xi_adj(j,dn)*exp(D)+xi_adj(j,dn)*(R_riskyassets_exp(:,2,dn)-exp(D))).^(1-gamma)).*H_comb(:,j,dn));       
    end

    %update H_der when no trade
    for j=1:Numsteps
        if j==1 
            H_der_NT(j,dn+1)=(H_NT(j+1,dn+1)-H_NT(j,dn+1))/(1/(Numsteps));
        elseif j==Numsteps
            H_der_NT(j,dn+1)=(H_NT(j,dn+1)-H_NT(j-1,dn+1))/(1/(Numsteps));
        else
            H_der_NT(j,dn+1)=(H_NT(j+1,dn+1)-H_NT(j-1,dn+1))/(2/(Numsteps));
        end
    end

    %finding optimal xi when opportunity to trade
    H_NT_T(1,dn+1)=min(H_NT(1:Numsteps,dn+1));

    for j=1:Numsteps
         if H_NT(j,dn+1)==H_NT_T(1,dn+1)
            xi_ster_TC(dn,1)=xi_adj(j,dn);
            consumption_ster_TC(dn,1)=parameter_estimates(j,dn,1);
            liquid_ster_TC(dn,1)=parameter_estimates(j,dn,2);
            H_comb_ster_TC(:,dn)=H_comb(:,j,dn);
         end
    end

    %correct for transactions costs when trade
    for j=1:Numsteps
       H_NT_T_with_TC(j,dn+1)=(consumption_ster_TC(dn,1))^(1-gamma)+beta^(1/freq)*mean(((((iota-xi_ster_TC(dn,1).*iota-tc*abs(xi_adj(j,dn).*iota-xi_ster_TC(dn,1).*iota)-...
       consumption_ster_TC(dn,1)*iota-LI(:,dn)).*(r_exp+liquid_ster_TC(dn,1)*(R_riskyassets_exp(:,1,dn)-r_exp))+...
       xi_ster_TC(dn,1)*exp(D)+xi_ster_TC(dn,1).*(R_riskyassets_exp(:,2,dn)-exp(D))).^(1-gamma)).*H_comb_ster_TC(:,dn)));  
         
       H_NT_T_with_TC_int(j,dn+1)=H_NT_T_with_TC(j,dn+1); %transaction cost when you enter are tc|0-xi_ster|
             if H_NT_T_with_TC(j,dn+1)>H_NT(j,dn+1)
             H_NT_T_with_TC(j,dn+1)=H_NT(j,dn+1);  %no-trade region: replace H_NT_T_with_TC with H_NT if H_NT_T_with_TC larger than H_NT
             end       
    end

    %find H_comb and H_comb_der
    for j=1:Numsteps
         for i=1:M
          H_comb(i,j,dn+1)=p_t(i,dn)*H_NT_T_with_TC(j,dn+1)+(1-p_t(i,dn))*H_NT(j,dn+1);
          H_comb_der(i,j,dn+1)=(1-p_t(i,dn))*H_der_NT(j,dn+1);
         end
    end
end

clear consumption_ster_TC xi_ster_TC liquid_ster_TC H_comb_ster_TC


%% solve liquid case

%define vectors and matrices
x0_trade=zeros(Numsteps_delta-1,T*freq,3); x0_trade_noCons=zeros(Numsteps_delta-1,T*freq,3); sum_square_residuals_liquid=zeros(Numsteps_delta,T*freq);
dc_exp_ret=zeros(Numsteps_delta-1,1); dc_exp_ret_exp=zeros(Numsteps_delta-1,1); termination_L=zeros(Numsteps,T*freq);

%define grid search to determine shadow costs of illiquidity 
for j=1:Numsteps_delta-1
    if lambda2<0.5
        dc_exp_ret(j,1)=1/freq*((j-1)/((1/sigma2*lambda2)*(Numsteps*9))); %solve liquid case for different decreases in expected returns 
    else
       dc_exp_ret(j,1)=1/freq*((j-1)/((1/sigma2*lambda2)*(Numsteps/2.5))); %solve liquid case for different decreases in expected returns  
    end
    dc_exp_ret_exp(j,1)=exp(-dc_exp_ret(j,1)); 
end

%start dynamic optimization problem 
lb = [0,-inf,-inf]; %lower bound
ub = [1,inf,inf]; %upper bound

for dn=1:T*freq
    
    %initial parameter values
    if ll<0.5
    x0_trade_initial = [1/(dn+1)*(1-mean(LI(:,dn))), mertonsol(1,1)/(1-mertonsol(2,1)), mertonsol(2,1)*(1-1/(dn+1)*(1-mean(LI(:,dn))))];
    else
    x0_trade_initial = [1/(dn+1)*(1-ll), mertonsol(1,1)/(1-mertonsol(2,1)), mertonsol(2,1)*(1-1/(dn+1)*(1-ll))];    
    end
    
    for j=1:Numsteps_delta-1
        fun_T = @(x) focs_T(x,beta,gamma,r_exp,R_riskyassets_exp(:,:,dn),dc_exp_ret_exp(j,1),H_T_liquid(j,dn),iota,LI(:,dn),1/freq); %identifying function to be solved  %initial values parameters of interest
        [sol_trade,resnorm,lambda,exitflag]=lsqnonlin(fun_T,x0_trade_initial,lb,ub,options); %applying fsolve to obtain optimal solution for xi, alpha_xi and theta_xi
        x0_trade(j,dn,:)=sol_trade; %solution solver for decrease in expected return at step j
        sum_square_residuals_liquid(j,dn)=resnorm; %obtain sum squared residuals
        termination_L(j,dn)=exitflag; %reason solver stopped

        %updating H-trade liquid case
        H_T_liquid(j,dn+1)=(x0_trade(j,dn,1))^(1-gamma)+beta^(1/freq)*mean(((iota-x0_trade(j,dn,1)*iota-x0_trade(j,dn,3)*iota-LI(:,dn)).*(r_exp+x0_trade(j,dn,2)*(R_riskyassets_exp(:,1,dn)-r_exp))+...
        x0_trade(j,dn,3)*(R_riskyassets_exp(:,2,dn).*dc_exp_ret_exp(j,1))).^(1-gamma).*H_T_liquid(j,dn));

        x0_trade_initial = x0_trade(j,dn,:); %update initial parameters
        
     end
end


%% shadow costs
% shadow costs is determined where
% min(H_NT(:,dn+1))==H_T_liquid(\delta,dn+1); see paper

%define vectors and matrices
H_trade_NT=zeros(T*freq+1,1); val=zeros(T*freq,1); idx=zeros(T*freq,1); shadow_costs=zeros(T*freq,1);

%find shadow costs for investor with horizon dn
for dn=1:T*freq
        H_trade_NT(dn+1,1)=min(H_NT_T_with_TC_int(1:Numsteps,dn+1));
        [val(dn,1),idx(dn,1)]=min(abs(H_T_liquid(:,dn+1)-H_trade_NT(dn+1,1)));
        shadow_costs(dn,1)=dc_exp_ret(idx(dn),1);
end
clear idx val

%annualize shadow costs
shadow_costs=shadow_costs*freq;


%% optimal policies illiquid and liquid case
%define vectors and matrices
consumption=zeros(Numsteps,T*freq); w_illiquid=zeros(Numsteps,T*freq); w_liquid=zeros(Numsteps,T*freq);
consumption_L=zeros(Numsteps_delta,T*freq); w_illiquid_L=zeros(Numsteps_delta,T*freq); w_liquid_L=zeros(Numsteps_delta,T*freq);
val=zeros(T*freq,1); idx=zeros(T*freq,1); consumption_ster=zeros(T*freq,1); xi_ster=zeros(T*freq,1); liquid_ster=zeros(T*freq,1);

%consumption and asset allocation illiquid case
for dn=1:T*freq
    for j=1:Numsteps
    consumption(j,dn)=parameter_estimates(j,dn,1);
    w_illiquid(j,dn)=xi_adj(j,dn)/(1-consumption(j,dn));    
    w_liquid(j,dn)=(1-xi_adj(j,dn)/(1-consumption(j,dn)-ll))*parameter_estimates(j,dn,2);
    end 
end

%optimal policies illiquid case
for dn=1:T*freq  
   [val(dn,1),idx(dn,1)]=min(H_NT_T_with_TC_int(1:Numsteps,dn+1));
    consumption_ster(dn,1)=parameter_estimates(idx(dn,1),dn,1);
    xi_ster(dn,1)=w_illiquid(idx(dn,1),dn);
    liquid_ster(dn,1)=w_liquid(idx(dn,1),dn);
end

%consumption and asset allocation liquid case
for dn=1:1:T*freq  
    for j=1:Numsteps_delta-1
        consumption_L(j,dn)=x0_trade(j,dn,1);
        w_illiquid_L(j,dn)=x0_trade(j,dn,3)/(1-consumption_L(j,dn)-mean(LI(:,dn)));
        w_liquid_L(j,dn)=(1-w_illiquid_L(j,dn))*x0_trade(j,dn,2);
    end
end

clear idx val


%% Graphs
%shadow costs
h=figure;
plot(1:T*freq,smooth(shadow_costs)*10000,'LineWidth',3)
xlim([1 T*freq])
xlabel('Horizon','FontSize',14)
ylabel('Shadow costs (bp)','FontSize',14)
grid on

set(gca,'FontSize',13)
set(h,'Units','Inches');
pos = get(h,'Position');
set(h,'PaperPositionMode','Auto','PaperUnits','Inches','PaperSize',[pos(3), pos(4)])









